import numpy as np
from PIL import Image
import sys
import matplotlib.pyplot as plt
from pathlib import Path
import os
import requests
import zipfile

def download_tutorial_07(target_file='tutorial_07_files.zip'):
    source_url = f'https://is.muni.cz/el/fi/jaro2024/PA228/um/exercises/tutorial_07/{target_file}'
    print(f'Downloading data from {source_url} to {target_file}')
    if not os.path.exists(target_file):
        print(f'\tDownloading ... {source_url}')
        content = requests.get(source_url).content
        with open(target_file, 'wb') as file:
            file.write(content)
    else:
         print(f'\tFile already downloaded')   
    folder_name = target_file.replace('.zip', '')
    if not os.path.exists(folder_name):
        print(f'\tExtracting to ... {folder_name}')
        with zipfile.ZipFile(target_file, 'r') as zip_ref:
            zip_ref.extractall(folder_name)
    else:
        print(f'\tFile already extracted')   

def plot_output_anchors(padded, anchors, torch_output_18, good_dets=None):
    if good_dets is None:
        good_dets = range(len(anchors))
    plt.rcParams['figure.figsize'] = [15, 10]
    fig = plt.figure()

    # Plot the anchor locations
    ax = fig.add_subplot(1, 3, 1)
    ax.imshow(np.array(padded))
    ax.scatter([anchors[i][0]*192 for i in good_dets], [anchors[i][1]*192 for i in good_dets], s=1)
    ax.set_title('Anchor locations')

    # Visualize the offsets from the SSD outputs. Notice how many bounding boxes gravitate towards palm centers
    ax = fig.add_subplot(1, 3, 2)
    ax.imshow(np.array(padded))
    ax.quiver([anchors[i][0]*192 for i in good_dets], [anchors[i][1]*192 for i in good_dets], [torch_output_18[0][i][0] for i in good_dets], [torch_output_18[0][i][1] for i in good_dets], angles='xy', scale_units='xy', scale=1)
    ax.set_title('Anchor offsets')

    # Visualize the sizes of the bounding boxes. Notice that every location has multiple, highly different sizes. (Aspect rations are almost always close to ~1:1)
    #ax = fig.add_subplot(1, 3, 3, projection='3d')
    ax = fig.add_subplot(1, 3, 3)
    ax.imshow(np.array(padded))
    ax.scatter([anchors[i][0]*192 + torch_output_18[0][i][0] for i in good_dets], [anchors[i][1]*192 + torch_output_18[0][i][1] for i in good_dets], s=1)
    ax.set_title('Final locations')
    ax.set_xlim([0, 192])
    ax.set_ylim([0, 192])
    ax.invert_yaxis()

    plt.show()

def plot_detections(detections, draw_padded, padded, w, h, title, draw_kp=False):
    for det in detections:
        xmin = int(det[2][0] * w)
        ymin = int(det[2][1] * h)
        xmax = int((det[2][0] + det[2][2]) * w)
        ymax = int((det[2][1] + det[2][3]) * h)
        draw_padded.rectangle(((xmin, ymin, xmax, ymax)), outline=(255, 0, 0))
        center_x = (xmin + xmax)/2
        center_y = (ymin + ymax)/2
        #draw_padded.ellipse((center_x-3, center_y-3, center_x+3, center_y+3), outline=(255, 0, 255))
        if draw_kp:
            for kp in det[3]:
                draw_padded.ellipse((int(kp[0]*w)-1, int(kp[1]*h)-1, int(kp[0]*w)+1, int(kp[1]*h)+1), outline=(0, 0, 255))

    plt.rcParams['figure.figsize'] = [5, 5]
    plt.imshow(np.array(padded))
    plt.title(title)
    
def plot_target_anchors(image_annot, anchors, target_coords, target_confs):
    plt.rcParams['figure.figsize'] = [15, 10]
    fig = plt.figure()

    # Plot the anchor locations
    ax = fig.add_subplot(1, 3, 1)
    ax.imshow(np.array(image_annot))
    ax.scatter([a[0]*192 for a in anchors], [a[1]*192 for a in anchors], s=1)
    ax.set_title('Anchor locations')

    # Visualize the offsets from the SSD outputs. Notice how many bounding boxes gravitate towards palm centers
    ax = fig.add_subplot(1, 3, 2)
    ax.imshow(np.array(image_annot))
    ax.quiver([a[0]*192 for a in anchors], [a[1]*192 for a in anchors], [b[0] for b in target_coords], [b[1] for b in target_coords], angles='xy', scale_units='xy', scale=1)
    ax.set_title('Anchor offsets')

    # Visualize the sizes of the bounding boxes. Notice that every location has multiple, highly different sizes. (Aspect rations are almost always close to ~1:1)
    ax = fig.add_subplot(1, 3, 3, projection='3d')
    ax.scatter([a[0]*192 for a in anchors], [a[1]*192 for a in anchors], [b for b in target_confs], c=[b for b in target_confs], cmap='plasma')
    ax.set_title('Confidence scores')
    ax.set_zlim([-5, 5])

    plt.show()
    

def ishow(img,
          cmap='viridis',
          title='',
          fig_size=(8,6),
          colorbar=False,
          interpolation='none'):
    ' Function `ishow` displays an image in a new window. '
    
    # colorbar only for 2D images
    if (len(img.shape) == 2) or (img.shape[2] == 1):
        colorbar=True
    
    extent = (0, img.shape[1], img.shape[0], 0)
    fig, ax = plt.subplots(figsize=fig_size)
    pcm = ax.imshow(img,
              extent=extent,
              cmap=cmap,
              interpolation=interpolation)
    
    ax.set_frame_on(False)
    plt.title(title)
    plt.tight_layout()
    if colorbar:
        fig.colorbar(pcm, orientation='vertical')
    plt.show()
    
# ------ MEDIAPIPE FUNCTIONS ------ 

def DetectionLetterboxRemoval(detection, w_og, h_og, letterbox_x, letterbox_y):
    new_det = [detection[0], detection[1], (), []]
    xmin, ymin, w, h = detection[2]
    x_ratio = w_og/(192-letterbox_x*2)
    y_ratio = h_og/(192-letterbox_y*2)
    w_ratio = 192/(192-letterbox_x*2)
    h_ratio = 192/(192-letterbox_y*2)
    
    new_xmin = ((192*xmin - letterbox_x) * x_ratio)/w_og
    new_w = w * w_ratio

    new_ymin = ((192*ymin - letterbox_y) * y_ratio)/h_og
    new_h = h * h_ratio
    new_det[2] = (new_xmin, new_ymin, new_w, new_h)
    for kp in detection[3]:
        new_det[3].append((((192*kp[0] - letterbox_x) * x_ratio)/w_og, ((192*kp[1] - letterbox_y) * y_ratio)/h_og))
    return new_det
        
target_angle = np.pi * 90.0 / 180.0
  
def NormalizedRadians(angle):
    return angle - 2 * np.pi * np.floor((angle - (-np.pi)) / (2 * np.pi))

def ComputeRotation(detection):
    x0 = detection[3][0][0]
    y0 = detection[3][0][1]
    x1 = detection[3][2][0]
    y1 = detection[3][2][1]
    return NormalizedRadians(target_angle - np.arctan2(-(y1 - y0), x1 - x0))

def RectFromBox(box):
    x_center = box[0] + box[2]/2
    y_center = box[1] + box[3]/2
    return ([x_center, y_center, box[2], box[3], 0.0])

def DetectionToNormalizedRect(detection):
    rect = RectFromBox(detection[2])
    return rect

def DetectionsToRects(detections):
    rects = []
    for det in detections:
        rect = DetectionToNormalizedRect(det)
        rot = ComputeRotation(det)
        rect[4] = rot
        rects.append(rect)
    return rects

def ImageToTensor(image, w, h):
    w0 = image.width
    h0 = image.height
    scale_w = w/w0
    scale_h = h/h0
    if w0 > h0:
        new_w = int(round(w0*scale_w))
        new_h = int(round(h0*scale_w))
    else:
        new_w = int(round(w0*scale_h))
        new_h = int(round(h0*scale_h))
    
    letterbox_x = int(round((w - new_w)/2))
    letterbox_y = int(round((h - new_h)/2))
    # correction
    new_w = w - 2*letterbox_x
    new_h = h - 2*letterbox_y
    
    image = image.resize((new_w, new_h))
    padded = Image.new(image.mode, (w, h), (0, 0, 0))
    
    if letterbox_x == 0 and letterbox_y == 0:
        padded = image
    elif letterbox_x == 0:
        padded.paste(image, (0, letterbox_y))
    elif letterbox_y == 0:
        padded.paste(image, (letterbox_x, 0))
    else: # Impossible
        padded.paste(image, (letterbox_x, letterbox_y))
    
    ret_im = np.array(padded)
    ret_im = ret_im / 255
    ret_im = np.moveaxis(ret_im, 2, 0)
    ret_im = np.expand_dims(ret_im, 0).astype(dtype=np.float32)
    return ret_im, padded, letterbox_x, letterbox_y

def LandmarkLetterboxRemoval(landmark, w_og, h_og, letterbox_x, letterbox_y):
    x, y, z = landmark
    x_ratio = w_og/(224-letterbox_x*2)
    y_ratio = h_og/(224-letterbox_y*2)

    new_x = ((224*x - letterbox_x) * x_ratio)/w_og
    new_y = ((224*y - letterbox_y) * y_ratio)/h_og
    new_z = z
    return (new_x, new_y, new_z)
        
        
        
min_suppression_threshold = 0.3

def OverlapSimilarity(det1, det2):
    #Only implement IoU
    x11, y11, w1, h1 = det1[2]
    x21, y21, w2, h2 = det2[2]
    x12, y12 = x11 + w1, y11 + h1
    x22, y22 = x21 + w2, y21 + h2
    area1 = (x12 - x11) * (y12 - y11)
    area2 = (x22 - x21) * (y22 - y21)
    if (min(x12, x22) - max(x11, x21)) > 0 and (min(y12, y22) - max(y11, y21)) > 0:
        intersection = (min(x12, x22) - max(x11, x21)) * (min(y12, y22) - max(y11, y21))
        if intersection < 0:
            return 0
        union = area1 + area2 - intersection
        return intersection/union
    return 0

def WeightedNonMaximumSuppression(indexed_scores, detections, output_detections, custom_metric=None):
    remained_indexed_scores = indexed_scores.copy()
    remained = []
    candidates = []
    output_detections.clear()
    while not len(remained_indexed_scores) == 0:
        original_indexed_scores_size = len(remained_indexed_scores)
        detection = detections[remained_indexed_scores[0][0]]
        remained.clear()
        candidates.clear()
        for indexed_score in remained_indexed_scores:
            second_det = detections[indexed_score[0]]
            if custom_metric is None:
                similarity = OverlapSimilarity(second_det, detection)
            else:
                similarity = custom_metric(second_det, detection)
                
            if similarity > min_suppression_threshold:
                candidates.append(indexed_score)
            else:
                remained.append(indexed_score)
        weighted_detection = detection
        if not len(candidates) == 0:
            num_keypoints = 7
            keypoints = [0]*num_keypoints*2
            w_xmin = 0.0
            w_ymin = 0.0
            w_xmax = 0.0
            w_ymax = 0.0
            total_score = 0.0
            for candidate in candidates:
                total_score += candidate[1]
                bbox = detections[candidate[0]]
                w_xmin += bbox[2][0] * candidate[1]
                w_ymin += bbox[2][1] * candidate[1]
                w_xmax += (bbox[2][0] + bbox[2][2]) * candidate[1]
                w_ymax += (bbox[2][1] + bbox[2][3]) * candidate[1]
                for i in range(len(bbox[3])):
                    keypoints[2*i] += bbox[3][i][0] * candidate[1]
                    keypoints[2*i + 1] += bbox[3][i][1] * candidate[1]
            new_x = w_xmin/total_score
            new_y = w_ymin/total_score
            new_w = (w_xmax-w_xmin)/total_score
            new_h = (w_ymax-w_ymin)/total_score
            weighted_detection[2] = (new_x, new_y, new_w, new_h)
            new_keypoints = []
            for i in range(len(bbox[3])):
                new_keypoints.append((keypoints[2*i]/total_score, keypoints[2*i+1]/total_score))
            weighted_detection[3] = new_keypoints
            output_detections.append(weighted_detection)
            if original_indexed_scores_size == len(remained):
                break
            else:
                remained_indexed_scores = remained.copy()
                    

def NonMaxSuppression(input_detections, custom_metric=None):
    # Remove non-max scoring labels from every detection
    indexed_scores = []
    for idx, det in enumerate(input_detections):
        indexed_scores.append((idx, det[0]))
    indexed_scores.sort(reverse=True, key=lambda elem: elem[1])
    # We could define the maximum number of detections here
    retained_detections = []
    WeightedNonMaximumSuppression(indexed_scores, input_detections, retained_detections, custom_metric)
    return retained_detections
    
    
rect_options = {
  'scale_x': 2.6,
  'scale_y': 2.6,
  'shift_y': -0.5,
  'square_long': True}
      
#Shift the rect towards the tip of the hand
def TransformNormalizedRect(rect, image_width, image_height):
    options = rect_options
    width = rect[2]
    height = rect[3]
    rotation = rect[4]
    
    if rotation == 0.0: #don't ask...
        rect[0] = rect[0] + width * rect_options.get('shift_x', 0.0)
        rect[1] = rect[1] + height * rect_options.get('shift_y', 0.0)
    else:
        x_shift = (image_width * width * rect_options.get('shift_x', 0.0) * np.cos(rotation) - image_height * height * rect_options.get('shift_y', 0.0) * np.sin(rotation)) / image_width
        y_shift = (image_width * width * rect_options.get('shift_x', 0.0) * np.sin(rotation) + image_height * height * rect_options.get('shift_y', 0.0) * np.cos(rotation)) / image_height
        rect[0] = rect[0] + x_shift
        rect[1] = rect[1] + y_shift
    if rect_options.get('square_long', False):
        long_side = max(width * image_width, height * image_height)
        width = long_side / image_width
        height = long_side / image_height
    else:
        short_side = min(width * image_width, height * image_height)
        width = short_side / image_width
        height = short_side / image_height
    rect[2] = width * rect_options.get('scale_x', 0.0)
    rect[3] = height * rect_options.get('scale_y', 0.0)
    return rect

def TransformNormalizedRects(rects, image_width, image_height):
    new_rects = []
    for rect in rects:
        new_rect = TransformNormalizedRect(rect, image_width, image_height)
        new_rects.append(new_rect)
    return new_rects

anchor_options = {
      'num_layers': 4,
      'min_scale': 0.1484375,
      'max_scale': 0.75,
      'input_size_width': 192,
      'input_size_height': 192,
      'anchor_offset_x': 0.5,
      'anchor_offset_y': 0.5,
      'strides': [8, 16, 16, 16],
      'aspect_ratios': [1.0],
      'fixed_anchor_size': True
}

def CalculateScale(min_scale, max_scale, stride_index, num_strides):
    if num_strides == 1:
        return (min_scale + max_scale) * 0.5
    else:
        return min_scale + (max_scale - min_scale) * 1.0 * stride_index / (num_strides - 1.0)

def GenerateAnchors():
    options = anchor_options
    num_layers = options['num_layers']
    layer_id = 0
    anchors = []
    while layer_id < num_layers:
        anchor_height = []
        anchor_width = []
        aspect_ratios = []
        scales = []
        last_same_stride_layer = layer_id
        while last_same_stride_layer < len(options['strides']) and options['strides'][last_same_stride_layer] == options['strides'][layer_id]:
            scale = CalculateScale(options['min_scale'], options['max_scale'], last_same_stride_layer, len(options['strides']))
            if (last_same_stride_layer == 0 and options.get('reduce_boxes_in_lowest_layer', False)):
                aspect_ratios.append(1.0)
                aspect_ratios.append(2.0)
                aspect_ratios.append(0.5)
                scales.append(0.1)
                scales.append(scale)
                scales.append(scale)
            else:
                for aspect_ratio_id in range(len(options['aspect_ratios'])):
                    aspect_ratios.append(options['aspect_ratios'][aspect_ratio_id])
                    scales.append(scale)    
                if options.get('interpolated_scale_aspect_ratio', 1.0) > 0.0:
                    scale_next = 1.0 if last_same_stride_layer == len(options['strides']) - 1 else CalculateScale(options['min_scale'], options['max_scale'], last_same_stride_layer + 1, len(options['strides']))
                    scales.append(np.sqrt(scale * scale_next))
                    aspect_ratios.append(options.get('interpolated_scale_aspect_ratio', 1.0))
            last_same_stride_layer += 1
        
        for i in range(len(aspect_ratios)):
            ratio_sqrts = np.sqrt(aspect_ratios[i])
            anchor_height.append(scales[i] / ratio_sqrts)
            anchor_width.append(scales[i] * ratio_sqrts)
        feature_map_height = 0
        feature_map_width = 0
        if options.get('feature_map_height_size', False):
            feature_map_height = options['feature_map_height_size'][layer_id]
            feature_map_width = options['feature_map_height_size'][layer_id]
        else:
            stride = options['strides'][layer_id]
            feature_map_height = np.ceil(1.0 * options['input_size_height'] / stride)
            feature_map_width = np.ceil(1.0 * options['input_size_width'] / stride)
        
        for y in range(int(feature_map_height)):
            for x in range(int(feature_map_width)):
                for anchor_id in range(len(anchor_height)):
                    x_center = (x + options['anchor_offset_x']) * 1.0 / feature_map_width
                    y_center = (y + options['anchor_offset_y']) * 1.0 / feature_map_height
                    if options['fixed_anchor_size']:
                        w = 1
                        h = 1
                    else:
                        w = anchor_width[anchor_id]
                        h = anchor_height[anchor_id]
                    anchors.append((x_center, y_center, w, h))
        layer_id = last_same_stride_layer
    return anchors

detection_options = {
      'num_classes': 1,
      'num_boxes': 2016,
      'num_coords': 18,
      'box_coord_offset': 0,
      'keypoint_coord_offset': 4,
      'num_keypoints': 7,
      'num_values_per_keypoint': 2,
      'sigmoid_score': True,
      'score_clipping_thresh': 100.0,
      'reverse_output_order': True,
      'x_scale': 192.0,
      'y_scale': 192.0,
      'w_scale': 192.0,
      'h_scale': 192.0,
      'min_score_thresh': 0.5
}

box_indices = [0, 1, 2, 3]

def DecodeBoxes(raw_boxes, anchors, boxes):
    options = detection_options
    for i in range(options.get('num_boxes')):
        box_offset = i * options.get('num_coords') + options.get('box_coord_offset')
        y_center = raw_boxes[0][i][options.get('box_coord_offset')]
        x_center = raw_boxes[0][i][options.get('box_coord_offset')+1]
        h = raw_boxes[0][i][options.get('box_coord_offset')+2]
        w = raw_boxes[0][i][options.get('box_coord_offset')+3]
        if options.get('reverse_output_order'):
            x_center = raw_boxes[0][i][options.get('box_coord_offset')]
            y_center = raw_boxes[0][i][options.get('box_coord_offset')+1]
            w = raw_boxes[0][i][options.get('box_coord_offset')+2]
            h = raw_boxes[0][i][options.get('box_coord_offset')+3]
        
        #print(x_center, y_center, w, h)
        x_center = x_center / options.get('x_scale') * anchors[i][2] + anchors[i][0]
        y_center = y_center / options.get('y_scale') * anchors[i][3] + anchors[i][1]

        h = h / options.get('h_scale') * anchors[i][3]
        w = w / options.get('w_scale') * anchors[i][2]

        ymin = y_center - h / 2.0
        xmin = x_center - w / 2.0
        ymax = y_center + h / 2.0
        xmax = x_center + w / 2.0
        boxes[i*options.get('num_coords')+0] = ymin
        boxes[i*options.get('num_coords')+1] = xmin
        boxes[i*options.get('num_coords')+2] = ymax
        boxes[i*options.get('num_coords')+3] = xmax
    
        if options.get('num_keypoints'):
            for k in range(options.get('num_keypoints')):
                offset = options.get('keypoint_coord_offset') + k * options.get('num_values_per_keypoint')
                keypoint_y = raw_boxes[0][i][offset]
                keypoint_x = raw_boxes[0][i][offset + 1]
                if options.get('reverse_output_order'):
                    keypoint_x = raw_boxes[0][i][offset]
                    keypoint_y = raw_boxes[0][i][offset + 1]
                boxes[i*options.get('num_coords') + offset] = keypoint_x / options.get('x_scale') * anchors[i][2] + anchors[i][0]
                boxes[i*options.get('num_coords') + offset+1] = keypoint_y / options.get('y_scale') * anchors[i][3] + anchors[i][1]

def IsClassIndexAllowed(class_index):
    return True

def ConvertToDetection(box_ymin, box_xmin, box_ymax, box_xmax, score, class_id, flip_vertically):
    detection = [score, class_id, (box_xmin, 1-box_ymin if flip_vertically else box_ymin, box_xmax-box_xmin, box_ymax-box_ymin), []]
    #score, class, (ox, oy, w, h), [(kx, ky)...]
    return detection

def ConvertToDetections(detection_boxes, detection_scores, detection_classes, output_detections):
    options = detection_options
    for i in range(options.get('num_boxes')):
        if options.get('max_results', 0) > 0 and len(output_detections) > options.get('max_results', 0):
            break
        if options.get('min_score_thresh', False) and detection_scores[i] < options.get('min_score_thresh', 0):
            continue
        if not IsClassIndexAllowed(detection_classes[i]):
            continue
        box_offset = i * options.get('num_coords')
        detection = ConvertToDetection(detection_boxes[box_offset + 0], #ymin
                                        detection_boxes[box_offset + 1], #xmin
                                        detection_boxes[box_offset + 2], #ymax
                                        detection_boxes[box_offset + 3], #xmax
                                        detection_scores[i], detection_classes[i], options.get('flip_vertically', False)) 
        if detection[2][2] < 0 or detection[2][3] < 0 or np.isnan(detection[2][2]) or np.isnan(detection[2][3]):
            continue
        if options.get('num_keypoints', 0) > 0:
            kp_id = 0
            while kp_id < options.get('num_keypoints') * options.get('num_values_per_keypoint'):
                keypoint_index = box_offset + options.get('keypoint_coord_offset') + kp_id
                keypoint = (detection_boxes[keypoint_index + 0], 1.0 - detection_boxes[keypoint_index + 1] if options.get('flip_vertically', False) else detection_boxes[keypoint_index + 1])
                detection[3].append(keypoint)
                kp_id += options.get('num_values_per_keypoint')
        output_detections.append(detection)

def FilterScores(raw_scores):
    options = detection_options
    good_dets = []
    for i in range(options.get('num_boxes')):
        score = raw_scores[0][i][0]
        score = -options.get('score_clipping_thresh') if score < -options.get('score_clipping_thresh') else score
        score = options.get('score_clipping_thresh') if score > options.get('score_clipping_thresh') else score
        score = 1.0 / (1.0 + np.exp(-score))
        if score > options.get('min_score_thresh', 0):
            good_dets.append(i)
    return good_dets

def TensorsToDetections(tensors, anchors):
    options = detection_options
    output_detections = []
    raw_boxes = tensors[0]
    raw_scores = tensors[1]
    boxes = [0]*(options.get('num_boxes') * options.get('num_coords'))
    DecodeBoxes(raw_boxes, anchors, boxes)
    detection_scores = [0]*options.get('num_boxes')
    detection_classes = [0]*options.get('num_boxes')
    for i in range(options.get('num_boxes')):
        class_id = -1
        max_score = -sys.float_info.min
        for score_idx in range(options.get('num_classes')):
            if(IsClassIndexAllowed(score_idx)):
                score = raw_scores[0][i][score_idx]
                if options.get('sigmoid_score'):
                    if options.get('score_clipping_thresh', False):
                        score = -options.get('score_clipping_thresh') if score < -options.get('score_clipping_thresh') else score
                        score = options.get('score_clipping_thresh') if score > options.get('score_clipping_thresh') else score
                    score = 1.0 / (1.0 + np.exp(-score))
                if max_score < score:
                    max_score = score
                    class_id = score_idx
        detection_scores[i] = max_score
        detection_classes[i] = class_id
    
    ConvertToDetections(boxes, detection_scores, detection_classes, output_detections)
    return output_detections
    